from openmdao.api import Group


class MphysGroup(Group):
    """
    A OpenMDAO class for automated promotion of MPhys variables
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.mphys_subsystems = []

    def mphys_add_subsystem(self, name, subsystem):
        """
        Adding an mphys subsystem will add the subsystem and then set the group
        to automatically promote the mphys tagged variables
        """
        subsystem = self.add_subsystem(name, subsystem)
        self.mphys_subsystems.append(subsystem)
        return subsystem

    def configure(self):
        """
        Promote the mphys-tagged variables of subsystems added by :func:`~MphysGroup.mphys_add_subsystem`
        """
        self._mphys_promote_coupling_variables()
        self._mphys_promote_inputs()
        self._mphys_promote_mesh_coordinates()
        self._mphys_promote_results()

    def _mphys_promote_by_tag(self, iotype, tag):
        for subsystem in self.mphys_subsystems:
            promoted = []
            tagged_variables = subsystem.get_io_metadata(iotypes=iotype,
                                                         metadata_keys=['tags'],
                                                         tags=tag)
            for val in tagged_variables.values():
                variable = val['prom_name']
                if variable not in promoted:
                    self.promotes(subsystem.name, any=[variable])
                    promoted.append(variable)

    def _mphys_promote_coupling_variables(self):
        self._mphys_promote_by_tag(['input', 'output'], 'mphys_coupling')

    def _mphys_promote_inputs(self):
        self._mphys_promote_by_tag('input', 'mphys_input')

    def _mphys_promote_mesh_coordinates(self):
        self._mphys_promote_by_tag(['input', 'output'], 'mphys_coordinates')

    def _mphys_promote_results(self):
        self._mphys_promote_by_tag('output', 'mphys_result')



from .mphys_group import MphysGroup
from .utils.directory_utils import cd

class Scenario(MphysGroup):
    """
    A group to represent a specific analysis condition or point of the MPhys
    multipoint groups.

    To make a Scenario for a particular type of multiphysics problem, subclass
    the Scenario, and implement the `initialize` and `setup` phases of the Group.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._post_subsystems = []

    def initialize(self):
        self.options.declare('run_directory',
                             default='',
                             types=str,
                             desc='Path in which to execute subsystems in this scenario group.' +
                                  ' The default of empty string will not change the directory.')

    def _mphys_scenario_setup(self):
        """
        This function is where specific scenarios populate pre-coupling, coupling,
        and post-coupling subsystems from builders
        """
        pass

    def setup(self):
        """
        The main setup call for all multiphysics scenarios.
        Multiphysics scenarios should implement setup-type operations in _mphys_scenario_setup().
        Adds the builder subsystems, then adds user-defined post subsystems.
        """
        self._mphys_scenario_setup()
        self._add_post_subsystems()

    def mphys_add_post_subsystem(self, name, subsystem,
                                 promotes_inputs=None,
                                 promotes_outputs=None,
                                 promotes=None):
        """
        Add a user-defined subsystem at the end of a Scenario.
        Tag variables with mphys tags to promote or use the optional promotes argument.

        Parameters
        ----------
        name: str
            Name of the subsystem
        subsystem: <System>
            The
        promotes: iter of (str or tuple), optional
            If None, variables will be promoted using mphys_* tags,
            else variables will be promoted by this input
        """

        # we hold onto these until the end of setup() b/c we want the scenario's
        # setup() to add the builder subsystems before adding these
        self._post_subsystems.append((name, subsystem, promotes_inputs, promotes_outputs, promotes))

    def _solve_nonlinear(self):
        with cd(self.options['run_directory']):
            return super()._solve_nonlinear()

    def _solve_linear(self, mode, rel_systems, scope_out=..., scope_in=...):
        with cd(self.options['run_directory']):
            return super()._solve_linear(mode, rel_systems, scope_out, scope_in)

    def _apply_nonlinear(self):
        with cd(self.options['run_directory']):
            return super()._apply_nonlinear()

    def _apply_linear(self, jac, rel_systems, mode, scope_out=None, scope_in=None):
        with cd(self.options['run_directory']):
            return super()._apply_linear(jac, rel_systems, mode, scope_out, scope_in)

    def _mphys_add_pre_coupling_subsystem_from_builder(self, name, builder, scenario_name=None):
        """
        If the builder has a precoupling subsystem, add it to the model.
        It is expected that is method is called during this scenario's`setup` phase.

        Parameters
        ----------
        name: str
            Name of the discipline
        builder: :class:`~mphys.builder.Builder`
            The discipline's builder object
        scenario_name: str or None
            Name of the scenario being setup (optional)
        """
        subsystem = builder.get_pre_coupling_subsystem(scenario_name)
        if subsystem is not None:
            self.mphys_add_subsystem(name+'_pre', subsystem)

    def _mphys_add_post_coupling_subsystem_from_builder(self, name, builder, scenario_name=None):
        """
        If the builder has a postcoupling subsystem, add it to the model.
        It is expected that is method is called during this scenario's`setup` phase.

        Parameters
        ----------
        name: str
            Name of the discipline
        builder: :class:`~mphys.builder.Builder`
            The discipline's builder object
        scenario_name: str or None
            Name of the scenario being setup (optional)
        """
        subsystem = builder.get_post_coupling_subsystem(scenario_name)
        if subsystem is not None:
            self.mphys_add_subsystem(name+'_post', subsystem)

    def _add_post_subsystems(self):
        for name, subsystem, promotes_inputs, promotes_outputs, promotes in self._post_subsystems:
            if self._no_promotes_specified(promotes_inputs, promotes_outputs, promotes):
                self.mphys_add_subsystem(name, subsystem)
            else:
                self.add_subsystem(name, subsystem,
                                   promotes_inputs=promotes_inputs,
                                   promotes_outputs=promotes_outputs,
                                   promotes=promotes)

    def _no_promotes_specified(self, promotes_inputs, promotes_outputs, promotes):
        return (promotes_inputs is None and
                promotes_outputs is None and
                promotes is None)



import numpy as np
import openmdao.api as om

class IntegratedSurfaceForces(om.ExplicitComponent):
    def setup(self):
        self.add_input('aoa',desc = 'angle of attack', units='rad',tags=['mphys_input'])
        self.add_input('yaw',desc = 'yaw angle',units='rad',tags=['mphys_input'])
        self.add_input('ref_area', val = 1.0,tags=['mphys_input'])
        self.add_input('moment_center',shape=3,tags=['mphys_input'])
        self.add_input('ref_length', val = 1.0,tags=['mphys_input'])
        self.add_input('q_inf', val = 1.0,tags=['mphys_input'])

        self.add_input('x_aero', shape_by_conn=True,
                                 distributed=True,
                                 desc = 'surface coordinates',
                                 tags=['mphys_coupling'])
        self.add_input('f_aero', shape_by_conn=True,
                                 distributed=True,
                                 desc = 'dimensional forces at nodes',
                                 tags=['mphys_coupling'])

        self.add_output('C_L', desc = 'Lift coefficient', tags=['mphys_result'])
        self.add_output('C_D', desc = 'Drag coefficient', tags=['mphys_result'])
        self.add_output('C_X', desc = 'X Force coefficient', tags=['mphys_result'])
        self.add_output('C_Y', desc = 'Y Force coefficient', tags=['mphys_result'])
        self.add_output('C_Z', desc = 'Z Force coefficient', tags=['mphys_result'])
        self.add_output('CM_X', desc = 'X Moment coefficient', tags=['mphys_result'])
        self.add_output('CM_Y', desc = 'Y Moment coefficient', tags=['mphys_result'])
        self.add_output('CM_Z', desc = 'Z Moment coefficient', tags=['mphys_result'])

        self.add_output('Lift', desc = 'Total Lift', tags=['mphys_result'])
        self.add_output('Drag', desc = 'Total Drag', tags=['mphys_result'])
        self.add_output('F_X', desc = 'Total X Force', tags=['mphys_result'])
        self.add_output('F_Y', desc = 'Total Y Force', tags=['mphys_result'])
        self.add_output('F_Z', desc = 'Total Z Force', tags=['mphys_result'])
        self.add_output('M_X', desc = 'Total X Moment', tags=['mphys_result'])
        self.add_output('M_Y', desc = 'Total Y Moment', tags=['mphys_result'])
        self.add_output('M_Z', desc = 'Total Z Moment', tags=['mphys_result'])

    def compute(self,inputs,outputs):
        aoa = inputs['aoa']
        yaw  = inputs['yaw']
        area = inputs['ref_area']
        q_inf = inputs['q_inf']
        xc = inputs['moment_center'][0]
        yc = inputs['moment_center'][1]
        zc = inputs['moment_center'][2]
        c = inputs['ref_length']

        x  = inputs['x_aero'][0::3]
        y  = inputs['x_aero'][1::3]
        z  = inputs['x_aero'][2::3]

        fx = inputs['f_aero'][0::3]
        fy = inputs['f_aero'][1::3]
        fz = inputs['f_aero'][2::3]

        fx_total = self.comm.allreduce(np.sum(fx))
        fy_total = self.comm.allreduce(np.sum(fy))
        fz_total = self.comm.allreduce(np.sum(fz))

        outputs['F_X'] = fx_total
        outputs['F_Y'] = fy_total
        outputs['F_Z'] = fz_total
        outputs['C_X'] = fx_total / (q_inf * area)
        outputs['C_Y'] = fy_total / (q_inf * area)
        outputs['C_Z'] = fz_total / (q_inf * area)

        outputs['Lift'] = -fx_total * np.sin(aoa) + fz_total * np.cos(aoa)
        outputs['Drag'] = ( fx_total * np.cos(aoa) * np.cos(yaw)
                          - fy_total * np.sin(yaw)
                          + fz_total * np.sin(aoa) * np.cos(yaw)
                          )

        outputs['C_L'] = outputs['Lift'] / (q_inf * area)
        outputs['C_D'] = outputs['Drag'] / (q_inf * area)

        m_x = self.comm.allreduce( np.dot(fz,(y-yc)) - np.dot(fy,(z-zc)))
        m_y = self.comm.allreduce(-np.dot(fz,(x-xc)) + np.dot(fx,(z-zc)))
        m_z = self.comm.allreduce( np.dot(fy,(x-xc)) - np.dot(fx,(y-yc)))

        outputs['M_X'] =  m_x
        outputs['M_Y'] =  m_y
        outputs['M_Z'] =  m_z

        outputs['CM_X'] = m_x / (q_inf * area * c)
        outputs['CM_Y'] = m_y / (q_inf * area * c)
        outputs['CM_Z'] = m_z / (q_inf * area * c)

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        aoa = inputs['aoa']
        yaw  = inputs['yaw']
        area = inputs['ref_area']
        q_inf = inputs['q_inf']
        xc = inputs['moment_center'][0]
        yc = inputs['moment_center'][1]
        zc = inputs['moment_center'][2]
        c = inputs['ref_length']

        x  = inputs['x_aero'][0::3]
        y  = inputs['x_aero'][1::3]
        z  = inputs['x_aero'][2::3]

        fx = inputs['f_aero'][0::3]
        fy = inputs['f_aero'][1::3]
        fz = inputs['f_aero'][2::3]

        fx_total = self.comm.allreduce(np.sum(fx))
        fy_total = self.comm.allreduce(np.sum(fy))
        fz_total = self.comm.allreduce(np.sum(fz))

        lift = -fx_total * np.sin(aoa) + fz_total * np.cos(aoa)
        drag = ( fx_total * np.cos(aoa) * np.cos(yaw)
               - fy_total * np.sin(yaw)
               + fz_total * np.sin(aoa) * np.cos(yaw)
               )

        m_x = self.comm.allreduce( np.dot(fz,(y-yc)) - np.dot(fy,(z-zc)))
        m_y = self.comm.allreduce(-np.dot(fz,(x-xc)) + np.dot(fx,(z-zc)))
        m_z = self.comm.allreduce( np.dot(fy,(x-xc)) - np.dot(fx,(y-yc)))

        if mode == 'fwd':
            if 'aoa' in d_inputs:
                daoa_rad = d_inputs['aoa']
                if 'Lift' in d_outputs or 'C_L' in d_outputs:
                    d_lift_d_aoa = ( - fx_total * np.cos(aoa) * daoa_rad
                                       - fz_total * np.sin(aoa) * daoa_rad )

                    if 'Lift' in d_outputs:
                        d_outputs['Lift'] += d_lift_d_aoa
                    if 'C_L' in d_outputs:
                        d_outputs['C_L'] += d_lift_d_aoa / (q_inf * area)
                if 'Drag' in d_outputs or 'C_D' in d_outputs:
                    d_drag_d_aoa = ( fx_total * (-np.sin(aoa) * daoa_rad) * np.cos(yaw)
                                     + fz_total * ( np.cos(aoa) * daoa_rad) * np.cos(yaw))
                    if 'Drag' in d_outputs:
                        d_outputs['Drag'] += d_drag_d_aoa
                    if 'C_D' in d_outputs:
                        d_outputs['C_D'] += d_drag_d_aoa / (q_inf * area)

            if 'yaw' in d_inputs:
                dyaw_rad = d_inputs['yaw']
                if 'Drag' in d_outputs or 'C_D' in d_outputs:
                    d_drag_d_yaw = ( fx_total * np.cos(aoa) * (-np.sin(yaw) * dyaw_rad)
                                    - fy_total * np.cos(yaw) * dyaw_rad
                                    + fz_total * np.sin(aoa) * (-np.sin(yaw) * dyaw_rad)
                                    )
                    if 'Drag' in d_outputs:
                        d_outputs['Drag'] += d_drag_d_yaw
                    if 'C_D' in d_outputs:
                        d_outputs['C_D'] += d_drag_d_yaw / (q_inf * area)

            if 'ref_area' in d_inputs:
                d_nondim = - d_inputs['ref_area'] / (q_inf * area**2.0)
                if 'C_X' in d_outputs:
                    d_outputs['C_X'] += fx_total * d_nondim
                if 'C_Y' in d_outputs:
                    d_outputs['C_Y'] += fy_total * d_nondim
                if 'C_Z' in d_outputs:
                    d_outputs['C_Z'] += fz_total * d_nondim
                if 'C_L' in d_outputs:
                    d_outputs['C_L'] += lift * d_nondim
                if 'C_D' in d_outputs:
                    d_outputs['C_D'] += drag * d_nondim
                if 'CM_X' in d_outputs:
                    d_outputs['CM_X'] += m_x * d_nondim / c
                if 'CM_X' in d_outputs:
                    d_outputs['CM_Y'] += m_y * d_nondim / c
                if 'CM_Z' in d_outputs:
                    d_outputs['CM_Z'] += m_z * d_nondim / c
            if 'moment_center' in d_inputs:
                dxc = d_inputs['moment_center'][0]
                dyc = d_inputs['moment_center'][1]
                dzc = d_inputs['moment_center'][2]
                if 'M_X' in d_outputs:
                    d_outputs['M_X'] += -fz_total * dyc + fy_total * dzc
                if 'M_Y' in d_outputs:
                    d_outputs['M_Y'] +=  fz_total * dxc - fx_total * dzc
                if 'M_Z' in d_outputs:
                    d_outputs['M_Z'] += -fy_total * dxc + fx_total * dyc
                if 'CM_X' in d_outputs:
                    d_outputs['CM_X'] += (-fz_total * dyc + fy_total * dzc) / (q_inf * area * c)
                if 'CM_Y' in d_outputs:
                    d_outputs['CM_Y'] += ( fz_total * dxc - fx_total * dzc) / (q_inf * area * c)
                if 'CM_Z' in d_outputs:
                    d_outputs['CM_Z'] += (-fy_total * dxc + fx_total * dyc) / (q_inf * area * c)

            if 'ref_length' in d_inputs:
                d_nondim = - d_inputs['ref_length'] / (q_inf * area * c**2.0)
                if 'CM_X' in d_outputs:
                    d_outputs['CM_X'] += m_x * d_nondim
                if 'CM_X' in d_outputs:
                    d_outputs['CM_Y'] += m_y * d_nondim
                if 'CM_Z' in d_outputs:
                    d_outputs['CM_Z'] += m_z * d_nondim

            if 'q_inf' in d_inputs:
                d_nondim = - d_inputs['q_inf'] / (q_inf**2.0 * area)
                if 'C_X' in d_outputs:
                    d_outputs['C_X'] += fx_total * d_nondim
                if 'C_Y' in d_outputs:
                    d_outputs['C_Y'] += fy_total * d_nondim
                if 'C_Z' in d_outputs:
                    d_outputs['C_Z'] += fz_total * d_nondim
                if 'C_L' in d_outputs:
                    d_outputs['C_L'] += lift * d_nondim
                if 'C_D' in d_outputs:
                    d_outputs['C_D'] += drag * d_nondim
                if 'CM_X' in d_outputs:
                    d_outputs['CM_X'] += m_x * d_nondim / c
                if 'CM_X' in d_outputs:
                    d_outputs['CM_Y'] += m_y * d_nondim / c
                if 'CM_Z' in d_outputs:
                    d_outputs['CM_Z'] += m_z * d_nondim / c

            if 'x_aero' in d_inputs:
                dx = d_inputs['x_aero'][0::3]
                dy = d_inputs['x_aero'][1::3]
                dz = d_inputs['x_aero'][2::3]
                if 'M_X' in d_outputs:
                    d_outputs['M_X'] +=  np.dot(fz,dy) - np.dot(fy,dz)
                if 'M_Y' in d_outputs:
                    d_outputs['M_Y'] += -np.dot(fz,dx) + np.dot(fx,dz)
                if 'M_Z' in d_outputs:
                    d_outputs['M_Z'] +=  np.dot(fy,dx) - np.dot(fx,dy)
                if 'CM_X' in d_outputs:
                    d_outputs['CM_X'] += ( np.dot(fz,dy) - np.dot(fy,dz)) / (q_inf * area * c)
                if 'CM_Y' in d_outputs:
                    d_outputs['CM_Y'] += (-np.dot(fz,dx) + np.dot(fx,dz)) / (q_inf * area * c)
                if 'CM_Z' in d_outputs:
                    d_outputs['CM_Z'] += ( np.dot(fy,dx) - np.dot(fx,dy)) / (q_inf * area * c)

            if 'f_aero' in d_inputs:
                dfx = d_inputs['f_aero'][0::3]
                dfy = d_inputs['f_aero'][1::3]
                dfz = d_inputs['f_aero'][2::3]
                dfx_total = np.sum(dfx)
                dfy_total = np.sum(dfy)
                dfz_total = np.sum(dfz)
                if 'F_X' in d_outputs:
                    d_outputs['F_X'] += dfx_total
                if 'F_Y' in d_outputs:
                    d_outputs['F_Y'] += dfy_total
                if 'F_Z' in d_outputs:
                    d_outputs['F_Z'] += dfz_total
                if 'C_X' in d_outputs:
                    d_outputs['C_X'] += dfx_total / (q_inf * area)
                if 'C_Y' in d_outputs:
                    d_outputs['C_Y'] += dfy_total / (q_inf * area)
                if 'C_Z' in d_outputs:
                    d_outputs['C_Z'] += dfz_total / (q_inf * area)
                if 'Lift' in d_outputs:
                    d_outputs['Lift'] += -dfx_total * np.sin(aoa) + dfz_total * np.cos(aoa)
                if 'Drag' in d_outputs:
                    d_outputs['Drag'] += ( dfx_total * np.cos(aoa) * np.cos(yaw)
                                         - dfy_total * np.sin(yaw)
                                         + dfz_total * np.sin(aoa) * np.cos(yaw)
                                         )
                if 'C_L' in d_outputs:
                    d_outputs['C_L'] += (-dfx_total * np.sin(aoa) + dfz_total * np.cos(aoa)) / (q_inf * area)
                if 'C_D' in d_outputs:
                    d_outputs['C_D'] += ( dfx_total * np.cos(aoa) * np.cos(yaw)
                                        - dfy_total * np.sin(yaw)
                                        + dfz_total * np.sin(aoa) * np.cos(yaw)
                                        ) / (q_inf * area)

                if 'M_X' in d_outputs:
                    d_outputs['M_X'] +=  np.dot(dfz,(y-yc)) - np.dot(dfy,(z-zc))
                if 'M_Y' in d_outputs:
                    d_outputs['M_Y'] += -np.dot(dfz,(x-xc)) + np.dot(dfx,(z-zc))
                if 'M_Z' in d_outputs:
                    d_outputs['M_Z'] +=  np.dot(dfy,(x-xc)) - np.dot(dfx,(y-yc))
                if 'CM_X' in d_outputs:
                    d_outputs['CM_X'] += ( np.dot(dfz,(y-yc)) - np.dot(dfy,(z-zc))) / (q_inf * area * c)
                if 'CM_Y' in d_outputs:
                    d_outputs['CM_Y'] += (-np.dot(dfz,(x-xc)) + np.dot(dfx,(z-zc))) / (q_inf * area * c)
                if 'CM_Z' in d_outputs:
                    d_outputs['CM_Z'] += ( np.dot(dfy,(x-xc)) - np.dot(dfx,(y-yc))) / (q_inf * area * c)

        elif mode == 'rev':
            if 'aoa' in d_inputs:
                if 'Lift' in d_outputs or 'C_L' in d_outputs:
                    d_lift = d_outputs['Lift'] if 'Lift' in d_outputs else 0.0
                    d_cl   = d_outputs['C_L']  if 'C_L'  in d_outputs else 0.0
                    d_inputs['aoa'] += ( - fx_total * np.cos(aoa)
                                         - fz_total * np.sin(aoa)
                                       ) * (d_lift + d_cl / (q_inf * area))

                if 'Drag' in d_outputs or 'C_D' in d_outputs:
                    d_drag = d_outputs['Drag'] if 'Drag' in d_outputs else 0.0
                    d_cd   = d_outputs['C_D']  if 'C_D'  in d_outputs else 0.0
                    d_inputs['aoa'] += ( fx_total * (-np.sin(aoa)) * np.cos(yaw)
                                       + fz_total * ( np.cos(aoa)) * np.cos(yaw)
                                       ) * (d_drag + d_cd / (q_inf * area))
            if 'yaw' in d_inputs:
                if 'Drag' in d_outputs or 'C_D' in d_outputs:
                    d_drag = d_outputs['Drag'] if 'Drag' in d_outputs else 0.0
                    d_cd   = d_outputs['C_D']  if 'C_D'  in d_outputs else 0.0
                    d_inputs['yaw'] += ( fx_total * np.cos(aoa) * (-np.sin(yaw))
                                       - fy_total * np.cos(yaw)
                                       + fz_total * np.sin(aoa) * (-np.sin(yaw))
                                       ) * (d_drag + d_cd / (q_inf * area))

            if 'ref_area' in d_inputs:
                d_nondim = - 1.0 / (q_inf * area**2.0)
                if 'C_X' in d_outputs:
                    d_inputs['ref_area'] += d_outputs['C_X'] * fx_total * d_nondim
                if 'C_Y' in d_outputs:
                    d_inputs['ref_area'] += d_outputs['C_Y'] * fy_total * d_nondim
                if 'C_Z' in d_outputs:
                    d_inputs['ref_area'] += d_outputs['C_Z'] * fz_total * d_nondim
                if 'C_L' in d_outputs:
                    d_inputs['ref_area'] += d_outputs['C_L'] * lift * d_nondim
                if 'C_D' in d_outputs:
                    d_inputs['ref_area'] += d_outputs['C_D'] * drag * d_nondim
                if 'CM_X' in d_outputs:
                    d_inputs['ref_area'] += d_outputs['CM_X'] * m_x * d_nondim / c
                if 'CM_X' in d_outputs:
                    d_inputs['ref_area'] += d_outputs['CM_Y'] * m_y * d_nondim / c
                if 'CM_Z' in d_outputs:
                    d_inputs['ref_area'] += d_outputs['CM_Z'] * m_z * d_nondim / c

            if 'moment_center' in d_inputs:
                if 'M_X' in d_outputs:
                    d_inputs['moment_center'][1] += -fz_total * d_outputs['M_X']
                    d_inputs['moment_center'][2] +=  fy_total * d_outputs['M_X']
                if 'M_Y' in d_outputs:
                    d_inputs['moment_center'][0] +=  fz_total * d_outputs['M_Y']
                    d_inputs['moment_center'][2] += -fx_total * d_outputs['M_Y']
                if 'M_Z' in d_outputs:
                    d_inputs['moment_center'][0] += -fy_total * d_outputs['M_Z']
                    d_inputs['moment_center'][1] +=  fx_total * d_outputs['M_Z']
                if 'CM_X' in d_outputs:
                    d_inputs['moment_center'][1] += -fz_total * d_outputs['CM_X'] / (q_inf * area * c)
                    d_inputs['moment_center'][2] +=  fy_total * d_outputs['CM_X'] / (q_inf * area * c)
                if 'CM_Y' in d_outputs:
                    d_inputs['moment_center'][0] +=  fz_total * d_outputs['CM_Y'] / (q_inf * area * c)
                    d_inputs['moment_center'][2] += -fx_total * d_outputs['CM_Y'] / (q_inf * area * c)
                if 'CM_Z' in d_outputs:
                    d_inputs['moment_center'][0] += -fy_total * d_outputs['CM_Z'] / (q_inf * area * c)
                    d_inputs['moment_center'][1] +=  fx_total * d_outputs['CM_Z'] / (q_inf * area * c)
            if 'ref_length' in d_inputs:
                d_nondim = - 1.0 / (q_inf * area * c**2.0)
                if 'CM_X' in d_outputs:
                    d_inputs['ref_length'] += m_x * d_nondim * d_outputs['CM_X']
                if 'CM_X' in d_outputs:
                    d_inputs['ref_length'] += m_y * d_nondim * d_outputs['CM_Y']
                if 'CM_Z' in d_outputs:
                    d_inputs['ref_length'] += m_z * d_nondim * d_outputs['CM_Z']

            if 'q_inf' in d_inputs:
                d_nondim = - 1.0 / (q_inf**2.0 * area)
                if 'C_X' in d_outputs:
                    d_inputs['q_inf'] += d_outputs['C_X'] * fx_total * d_nondim
                if 'C_Y' in d_outputs:
                    d_inputs['q_inf'] += d_outputs['C_Y'] * fy_total * d_nondim
                if 'C_Z' in d_outputs:
                    d_inputs['q_inf'] += d_outputs['C_Z'] * fz_total * d_nondim
                if 'C_L' in d_outputs:
                    d_inputs['q_inf'] += d_outputs['C_L'] * lift * d_nondim
                if 'C_D' in d_outputs:
                    d_inputs['q_inf'] += d_outputs['C_D'] * drag * d_nondim
                if 'CM_X' in d_outputs:
                    d_inputs['q_inf'] += d_outputs['CM_X'] * m_x * d_nondim / c
                if 'CM_X' in d_outputs:
                    d_inputs['q_inf'] += d_outputs['CM_Y'] * m_y * d_nondim / c
                if 'CM_Z' in d_outputs:
                    d_inputs['q_inf'] += d_outputs['CM_Z'] * m_z * d_nondim / c

            if 'x_aero' in d_inputs:
                nondim = 1.0 / (q_inf * area * c)
                dm_x = d_outputs['M_X'] if 'M_X' in d_outputs else 0.0
                dm_y = d_outputs['M_Y'] if 'M_Y' in d_outputs else 0.0
                dm_z = d_outputs['M_Z'] if 'M_Z' in d_outputs else 0.0
                dcm_x = d_outputs['CM_X']*nondim if 'CM_X' in d_outputs else 0.0
                dcm_y = d_outputs['CM_Y']*nondim if 'CM_Y' in d_outputs else 0.0
                dcm_z = d_outputs['CM_Z']*nondim if 'CM_Z' in d_outputs else 0.0
                d_inputs['x_aero'][0::3] += -fz * (dm_y + dcm_y) + fy * (dm_z + dcm_z)
                d_inputs['x_aero'][1::3] +=  fz * (dm_x + dcm_x) - fx * (dm_z + dcm_z)
                d_inputs['x_aero'][2::3] += -fy * (dm_x + dcm_x) + fx * (dm_y + dcm_y)

            if 'f_aero' in d_inputs:
                if 'F_X' in d_outputs:
                    d_inputs['f_aero'][0::3] += d_outputs['F_X']
                if 'F_Y' in d_outputs:
                    d_inputs['f_aero'][1::3] += d_outputs['F_Y']
                if 'F_Z' in d_outputs:
                    d_inputs['f_aero'][2::3] += d_outputs['F_Z']
                if 'C_X' in d_outputs:
                    d_inputs['f_aero'][0::3] += d_outputs['C_X'] / (q_inf * area)
                if 'C_Y' in d_outputs:
                    d_inputs['f_aero'][1::3] += d_outputs['C_Y'] / (q_inf * area)
                if 'C_Z' in d_outputs:
                    d_inputs['f_aero'][2::3] += d_outputs['C_Z'] / (q_inf * area)
                if 'Lift' in d_outputs:
                    d_inputs['f_aero'][0::3] += -np.sin(aoa) * d_outputs['Lift']
                    d_inputs['f_aero'][2::3] +=  np.cos(aoa) * d_outputs['Lift']
                if 'Drag' in d_outputs:
                    d_inputs['f_aero'][0::3] +=  np.cos(aoa) * np.cos(yaw) * d_outputs['Drag']
                    d_inputs['f_aero'][1::3] += -np.sin(yaw) * d_outputs['Drag']
                    d_inputs['f_aero'][2::3] +=  np.sin(aoa) * np.cos(yaw) * d_outputs['Drag']
                if 'C_L' in d_outputs:
                    d_inputs['f_aero'][0::3] += -np.sin(aoa) * d_outputs['C_L'] / (q_inf * area)
                    d_inputs['f_aero'][2::3] +=  np.cos(aoa) * d_outputs['C_L'] / (q_inf * area)
                if 'C_D' in d_outputs:
                    d_inputs['f_aero'][0::3] +=  np.cos(aoa) * np.cos(yaw) * d_outputs['C_D'] / (q_inf * area)
                    d_inputs['f_aero'][1::3] += -np.sin(yaw) * d_outputs['C_D'] / (q_inf * area)
                    d_inputs['f_aero'][2::3] +=  np.sin(aoa) * np.cos(yaw) * d_outputs['C_D'] / (q_inf * area)

                if 'M_X' in d_outputs:
                    d_inputs['f_aero'][1::3] += -(z-zc) * d_outputs['M_X']
                    d_inputs['f_aero'][2::3] +=  (y-yc) * d_outputs['M_X']
                if 'M_Y' in d_outputs:
                    d_inputs['f_aero'][0::3] +=  (z-zc) * d_outputs['M_Y']
                    d_inputs['f_aero'][2::3] += -(x-xc) * d_outputs['M_Y']
                if 'M_Z' in d_outputs:
                    d_inputs['f_aero'][0::3] += -(y-yc) * d_outputs['M_Z']
                    d_inputs['f_aero'][1::3] +=  (x-xc) * d_outputs['M_Z']
                if 'CM_X' in d_outputs:
                    d_inputs['f_aero'][1::3] += -(z-zc) * d_outputs['CM_X'] / (q_inf * area * c)
                    d_inputs['f_aero'][2::3] +=  (y-yc) * d_outputs['CM_X'] / (q_inf * area * c)
                if 'CM_Y' in d_outputs:
                    d_inputs['f_aero'][0::3] +=  (z-zc) * d_outputs['CM_Y'] / (q_inf * area * c)
                    d_inputs['f_aero'][2::3] += -(x-xc) * d_outputs['CM_Y'] / (q_inf * area * c)
                if 'CM_Z' in d_outputs:
                    d_inputs['f_aero'][0::3] += -(y-yc) * d_outputs['CM_Z'] / (q_inf * area * c)
                    d_inputs['f_aero'][1::3] +=  (x-xc) * d_outputs['CM_Z'] / (q_inf * area * c)

def check_integrated_surface_force_partials():
    nnodes = 3
    prob = om.Problem()
    ivc = om.IndepVarComp()
    ivc.add_output('aoa',val=45.0, units='deg')
    ivc.add_output('yaw',val=135.0, units='deg')
    ivc.add_output('ref_area',val=0.2)
    ivc.add_output('moment_center',shape=3,val=np.zeros(3))
    ivc.add_output('ref_length', val = 3.0)
    ivc.add_output('q_inf',val=10.0)
    ivc.add_output('x_aero',shape=3*nnodes,val=np.random.rand(3*nnodes),distributed=True)
    ivc.add_output('f_aero',shape=3*nnodes,val=np.random.rand(3*nnodes),distributed=True)
    prob.model.add_subsystem('ivc',ivc,promotes_outputs=['*'])
    prob.model.add_subsystem('forces',IntegratedSurfaceForces(),
                                      promotes_inputs=['*'])

    prob.setup(force_alloc_complex=True)
    prob.run_model()
    prob.check_partials(compact_print=True, method='cs')

if __name__ == '__main__':
    check_integrated_surface_force_partials()



import openmdao.api as om
from .scenario import Scenario
from .coupling_group import CouplingGroup


class ScenarioAeropropulsive(Scenario):
    def initialize(self):
        """
        A class to perform an aeropropulsive case.
        The Scenario will add the aerodynamic and propulsion builders' precoupling subsystem,
        the coupling subsystem, and the postcoupling subsystem.
        """
        super().initialize()

        self.options.declare("aero_builder", recordable=False, desc="The MPhys builder for the aerodynamic solver")
        self.options.declare("prop_builder", recordable=False, desc="The MPhys builder for the propulsion model")
        self.options.declare(
            "balance_builder", recordable=False, desc="The MPhys builder for the balance group", default=None
        )
        self.options.declare(
            "in_MultipointParallel",
            default=False,
            desc="Set to `True` if adding this scenario inside a MultipointParallel Group.",
        )
        self.options.declare(
            "geometry_builder", default=None, recordable=False, desc="The optional MPhys builder for the geometry"
        )

    def _mphys_scenario_setup(self):
        aero_builder = self.options["aero_builder"]
        prop_builder = self.options["prop_builder"]
        balance_builder = self.options["balance_builder"]
        geometry_builder = self.options["geometry_builder"]

        if self.options["in_MultipointParallel"]:
            self._mphys_initialize_builders(aero_builder, prop_builder, geometry_builder)
            self._mphys_add_mesh_and_geometry_subsystems(aero_builder, prop_builder, geometry_builder)

        self._mphys_add_pre_coupling_subsystem_from_builder("aero", aero_builder, self.name)
        self._mphys_add_pre_coupling_subsystem_from_builder("prop", prop_builder, self.name)

        coupling_group = CouplingAeropropulsive(
            aero_builder=aero_builder, prop_builder=prop_builder, balance_builder=balance_builder, scenario_name=self.name
        )
        self.mphys_add_subsystem("coupling", coupling_group)

        self._mphys_add_post_coupling_subsystem_from_builder("aero", aero_builder, self.name)
        self._mphys_add_post_coupling_subsystem_from_builder("prop", prop_builder, self.name)

    def _mphys_initialize_builders(self, aero_builder, prop_builder, geometry_builder):
        aero_builder.initialize(self.comm)
        prop_builder.initialize(self.comm)
        if geometry_builder is not None:
            geometry_builder.initialize(self.comm)

    def _mphys_add_mesh_and_geometry_subsystems(self, aero_builder, prop_builder, geometry_builder):

        if geometry_builder is None:
            self.mphys_add_subsystem("aero_mesh", aero_builder.get_mesh_coordinate_subsystem(self.name))
            # self.mphys_add_subsystem("prop_mesh", prop_builder.get_mesh_coordinate_subsystem(self.name))
        else:
            self.add_subsystem("aero_mesh", aero_builder.get_mesh_coordinate_subsystem(self.name))
            # the propulsion model does not need a mesh with pycycle
            # self.add_subsystem("prop_mesh", prop_builder.get_mesh_coordinate_subsystem(self.name))
            self.mphys_add_subsystem("geometry", geometry_builder.get_mesh_coordinate_subsystem(self.name))
            self.connect("aero_mesh.x_aero0", "geometry.x_aero_in")
            # the propulsion model does not need a mesh with pycycle
            # self.connect("prop_mesh.x_prop0", "geometry.x_prop_in")

    def mphys_make_aeroprop_conn(self, aero2prop_conn, prop2aero_conn):
        # TODO automate this with mphys_result or mphys_coupling tags

        # make the connections
        for k, v in aero2prop_conn.items():
            self.connect("coupling.aero.%s" % k, "coupling.prop.%s" % v)
        for k, v in prop2aero_conn.items():
            self.connect("coupling.prop.%s" % k, "coupling.aero.%s" % v)


class CouplingAeropropulsive(CouplingGroup):
    """
    The standard aeropropulsive coupling problem.
    """

    def initialize(self):
        self.options.declare("aero_builder", recordable=False)
        self.options.declare("prop_builder", recordable=False)
        self.options.declare("balance_builder", recordable=False, default=None)
        self.options.declare("scenario_name", recordable=True, default=None)

    def setup(self):
        aero_builder = self.options["aero_builder"]
        prop_builder = self.options["prop_builder"]
        balance_builder = self.options["balance_builder"]
        scenario_name = self.options["scenario_name"]

        aero = aero_builder.get_coupling_group_subsystem(scenario_name)
        prop = prop_builder.get_coupling_group_subsystem(scenario_name)

        self.mphys_add_subsystem("aero", aero)
        self.mphys_add_subsystem("prop", prop)

        if balance_builder is not None:
            balance = balance_builder.get_coupling_group_subsystem(scenario_name)
            self.mphys_add_subsystem("balance", balance)

        self.nonlinear_solver = om.NonlinearBlockGS(maxiter=25, iprint=2, atol=1e-8, rtol=1e-8)
        self.linear_solver = om.LinearBlockGS(maxiter=25, iprint=2, atol=1e-8, rtol=1e-8)



import os

class cd:
    def __init__(self, new_path: str):
        """
        Context manager for changing the current working directory.
        If an empty string is provided, the directory will not change.
        """
        if not new_path:
            new_path = os.getcwd()
        self.new_path = os.path.expanduser(new_path)

    def __enter__(self):
        self.saved_path = os.getcwd()
        os.chdir(self.new_path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_path)



#!/usr/bin/env python
from .builder import Builder
from .multipoint import Multipoint, MultipointParallel
from .distributed_converter import DistributedConverter, DistributedVariableDescription
from .mask_converter import MaskedConverter, UnmaskedConverter, MaskedVariableDescription



import numpy as np
import openmdao.api as om


class MaskedVariableDescription:
    """
    Attributes of a variable for conversion from masked to unmasked
    or unmasked to masked
    """

    def __init__(self, name: str, shape: tuple, tags=None):
        self.name = name
        self.shape = shape
        self.tags = tags


class MaskedConverter(om.ExplicitComponent):
    """
    An ExplicitComponent used to filter out a predefined set of indices from a larger input array.
    This is useful in cases, for instance, where it desired to prevent certain fea nodes from participating
    in the load and displacement transfers in an aerostructural scenario.

    The masking operation breaks down to the python assignment:
    masked_vector = unmasked_vector[mask_indices]
    """

    def initialize(self):
        self.options.declare(
            'input',
            desc='MaskedVariableDescription object of input that will be masked')
        self.options.declare(
            'output',
            desc='MaskedVariableDescription object of masked output')
        self.options.declare(
            'init_output', default=1.0,
            desc='initail value of the ouput. The default value matches the default value for val in add_output')
        self.options.declare(
            'distributed', default=False,
            desc='Flag to determine if the inputs and outputs should be distributed arrays')
        self.options.declare(
            'mask',
            desc='masking array to apply to vectors. Contains boolean flags '
                 'indicating which indices should be included (True) or masked (False)')

    def setup(self):
        distributed = self.options['distributed']
        input = self.options['input']
        output = self.options['output']
        mask = self.options['mask']

        self.add_input(input.name, shape=input.shape, tags=input.tags, distributed=distributed)

        if isinstance(output, list):
            if len(output) != len(mask):
                raise ValueError("Output length and mask length not equal")
            for i in range(len(output)):
                self.add_output(output[i].name, shape=output[i].shape, tags=output[i].tags, val=self.options['init_output'], distributed=distributed)
        else:
            self.add_output(output.name, shape=output.shape, tags=output.tags, val=self.options['init_output'], distributed=distributed)

    def compute(self, inputs, outputs):
        input = self.options['input']
        mask = self.options['mask']
        output = self.options['output']

        if isinstance(output, list):
            for i in range(len(output)):
                outputs[output[i].name] = inputs[input.name][mask[i]]
        else:
            outputs[output.name] = inputs[input.name][mask]

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        input = self.options['input']
        output = self.options['output']
        mask = self.options['mask']

        if isinstance(output, list):
            for i in range(len(output)):
                if mode == 'fwd':
                    if input.name in d_inputs and output[i].name in d_outputs:
                        d_outputs[output[i].name] += d_inputs[input.name][mask[i]]

                if mode == 'rev':
                    if input.name in d_inputs and output[i].name in d_outputs:
                        d_inputs[input.name][mask[i]] += d_outputs[output[i].name]
        else:
            if mode == 'fwd':
                if input.name in d_inputs and output.name in d_outputs:
                    d_outputs[output.name] += d_inputs[input.name][mask]

            if mode == 'rev':
                if input.name in d_inputs and output.name in d_outputs:
                    d_inputs[input.name][mask] += d_outputs[output.name]

class UnmaskedConverter(om.ExplicitComponent):
    """
    An ExplicitComponent that undoes the procedure of the MaskedConverter component.
    This companent takes an already masked vector, inserts missing indices,
    and gives back a vector of the full unmasked size. The user can optionally set
    the default values to be inserted for the inserted indices (default is 0.0).

    The unmasking operation breaks down to the python assignment:
    unmasked_vector[mask_indices] = masked_vector
    """

    def initialize(self):
        self.options.declare(
            'input',
            desc='MaskedVariableDescription object of input that will be unmasked')
        self.options.declare(
            'output',
            desc='MaskedVariableDescription object of unmasked output')
        self.options.declare(
            'default_values', default=0.0,
            desc='default values for masked indices in output vector')
        self.options.declare(
            'distributed', default=False,
            desc='Flag to determine if the inputs and outputs should be distributed arrays')
        self.options.declare(
            'mask',
            desc='masking array to apply to vectors. Contains boolean flags '
                 'indicating which indices should be included (True) or masked (False)')

    def setup(self):
        distributed = self.options['distributed']
        input = self.options['input']
        output = self.options['output']
        mask = self.options['mask']

        if isinstance(input, list):
            if len(input) != len(mask):
                raise ValueError("Input length and mask length not equal")

            for i in range(len(input)-1):
                for j in range(i+1, len(input)):
                    if (np.any(np.logical_and(mask[i], mask[j]))):
                        raise RuntimeWarning("Overlapping masking arrays, values will conflict.")

            for i in range(len(input)):
                self.add_input(input[i].name, shape=input[i].shape, tags=input[i].tags, distributed=distributed)
        else:
            self.add_input(input.name, shape=input.shape, tags=input.tags, distributed=distributed)

        self.add_output(output.name, shape=output.shape, tags=output.tags, distributed=distributed)

    def compute(self, inputs, outputs):
        input = self.options['input']
        output = self.options['output']
        mask = self.options['mask']
        def_vals = self.options['default_values']
        outputs[output.name][:] = def_vals

        if isinstance(input, list):
            for i in range(len(input)):
                outputs[output.name][mask[i]] = inputs[input[i].name]
        else:
            outputs[output.name][mask] = inputs[input.name]

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        input = self.options['input']
        output = self.options['output']
        mask = self.options['mask']

        if isinstance(input, list):
            for i in range(len(input)):
                if mode == 'fwd':
                    if input[i].name in d_inputs and output.name in d_outputs:
                        d_outputs[output.name][mask[i]] += d_inputs[input[i].name]

                if mode == 'rev':
                    if input[i].name in d_inputs and output.name in d_outputs:
                        d_inputs[input[i].name] += d_outputs[output.name][mask[i]]
        else:
            if mode == 'fwd':
                if input.name in d_inputs and output.name in d_outputs:
                    d_outputs[output.name][mask] += d_inputs[input.name]

            if mode == 'rev':
                if input.name in d_inputs and output.name in d_outputs:
                    d_inputs[input.name] += d_outputs[output.name][mask]



import openmdao.api as om
from .coupling_group import CouplingGroup
from .geo_disp import GeoDisp


class CouplingAeroThermal(CouplingGroup):
    """
    The standard aerothermal coupling problem.
    """

    def initialize(self):
        self.options.declare('aero_builder', recordable=False)
        self.options.declare('thermal_builder', recordable=False)
        self.options.declare('thermalxfer_builder', recordable=False)
        self.options.declare("scenario_name", recordable=True, default=None)

    def setup(self):
        aero_builder = self.options['aero_builder']
        
        thermal_builder = self.options['thermal_builder']
        
        thermalxfer_builder = self.options['thermalxfer_builder']
        
        scenario_name = self.options['scenario_name']

        heat_xfer, temp_xfer = thermalxfer_builder.get_coupling_group_subsystem(scenario_name)
        
        aero = aero_builder.get_coupling_group_subsystem(scenario_name)
        thermal = thermal_builder.get_coupling_group_subsystem(scenario_name)

        # geo_disp = GeoDisp(number_of_nodes=aero_builder.get_number_of_nodes())

        self.mphys_add_subsystem('thermal', thermal)
        self.mphys_add_subsystem('temp_xfer', temp_xfer)
        self.mphys_add_subsystem('aero', aero)
        self.mphys_add_subsystem('heat_xfer', heat_xfer)

        # add the defualt iterative solvers
        self.nonlinear_solver = om.NonlinearBlockGS(maxiter=25, iprint=2,
                                                    atol=1e-8, rtol=1e-8,
                                                    use_aitken=True)
        self.linear_solver = om.LinearBlockGS(maxiter=25, iprint=2,
                                              atol=1e-8, rtol=1e-8,
                                              use_aitken=True)



import openmdao.api as om

class GeoDisp(om.ExplicitComponent):
    """
    This component adds the aerodynamic
    displacements to the geometry-changed aerodynamic surface
    """
    def initialize(self):
        self.options.declare('number_of_nodes')

    def setup(self):
        nnodes = self.options['number_of_nodes']
        local_size = nnodes * 3

        self.add_input('x_aero0', shape_by_conn=True,
                                  distributed=True,
                                  desc='aerodynamic surface with geom changes',
                                  tags=['mphys_coordinates'])
        self.add_input('u_aero',  shape_by_conn=True,
                                  distributed=True,
                                  desc='aerodynamic surface displacements',
                                  tags=['mphys_coupling'])

        self.add_output('x_aero', shape=local_size,
                                  distributed=True,
                                  desc='deformed aerodynamic surface',
                                  tags=['mphys_coupling'])

    def compute(self,inputs,outputs):
        outputs['x_aero'] = inputs['x_aero0'] + inputs['u_aero']

    def compute_jacvec_product(self,inputs,d_inputs,d_outputs,mode):
        if mode == 'fwd':
            if 'x_aero' in d_outputs:
                if 'x_aero0' in d_inputs:
                    d_outputs['x_aero'] += d_inputs['x_aero0']
                if 'u_aero' in d_inputs:
                    d_outputs['x_aero'] += d_inputs['u_aero']
        if mode == 'rev':
            if 'x_aero' in d_outputs:
                if 'x_aero0' in d_inputs:
                    d_inputs['x_aero0'] += d_outputs['x_aero']
                if 'u_aero' in d_inputs:
                    d_inputs['u_aero']  += d_outputs['x_aero']



from mphys.scenario import Scenario
from mphys.coupling_aerostructural import CouplingAeroStructural


class ScenarioAeroStructural(Scenario):
    def initialize(self):
        """
        A class to perform a single discipline aerodynamic case.
        The Scenario will add the aerodynamic builder's precoupling subsystem,
        the coupling subsystem, and the postcoupling subsystem.
        """
        super().initialize()

        self.options.declare(
            "aero_builder",
            recordable=False,
            desc="The MPhys builder for the aerodynamic solver",
        )
        self.options.declare(
            "struct_builder",
            recordable=False,
            desc="The MPhys builder for the structural solver",
        )
        self.options.declare(
            "ldxfer_builder",
            recordable=False,
            desc="The MPhys builder for the load and displacement transfer",
        )
        self.options.declare(
            "in_MultipointParallel",
            default=False,
            desc="Set to `True` if adding this scenario inside a MultipointParallel Group.",
        )
        self.options.declare(
            "geometry_builder",
            default=None,
            recordable=False,
            desc="The optional MPhys builder for the geometry",
        )
        self.options.declare(
            "coupling_group_type",
            default="full_coupling",
            desc='Limited flexibility for coupling group type to accomodate flutter about jig shape or DLM where coupling group can be skipped: ["full_coupling", "aerodynamics_only", None]',
        )
        self.options.declare(
            "pre_coupling_order",
            default=["aero", "struct", "ldxfer"],
            recordable=False,
            desc="The order of the pre coupling subsystems",
        )
        self.options.declare(
            "post_coupling_order",
            default=["ldxfer", "aero", "struct"],
            recordable=False,
            desc="The order of the post coupling subsystems",
        )

    def _mphys_scenario_setup(self):
        if self.options["in_MultipointParallel"]:
            self._mphys_initialize_builders()
            self._mphys_add_mesh_and_geometry_subsystems()

        self._mphys_add_pre_coupling_subsystems()
        self._mphys_add_coupling_group()
        self._mphys_add_post_coupling_subsystems()

    def _mphys_check_coupling_order_inputs(self, given_options):
        valid_options = ["aero", "struct", "ldxfer"]

        length = len(given_options)
        if length > 3:
            raise ValueError(
                f"Specified too many items in the pre/post coupling order list, len={length}"
            )

        for option in given_options:
            if option not in valid_options:
                raise ValueError(
                    f"""Unknown pre/post order option: {option}. valid options are ["{'", "'.join(valid_options)}"]"""
                )

    def _mphys_add_pre_coupling_subsystems(self):
        self._mphys_check_coupling_order_inputs(self.options["pre_coupling_order"])
        for discipline in self.options["pre_coupling_order"]:
            self._mphys_add_pre_coupling_subsystem_from_builder(
                discipline, self.options[f"{discipline}_builder"], self.name
            )

    def _mphys_add_coupling_group(self):
        if self.options["coupling_group_type"] == "full_coupling":
            coupling_group = CouplingAeroStructural(
                aero_builder=self.options["aero_builder"],
                struct_builder=self.options["struct_builder"],
                ldxfer_builder=self.options["ldxfer_builder"],
                scenario_name=self.name,
            )
            self.mphys_add_subsystem("coupling", coupling_group)

        elif self.options["coupling_group_type"] == "aerodynamics_only":
            aero = self.options["aero_builder"].get_coupling_group_subsystem(self.name)
            self.mphys_add_subsystem("aero", aero)

    def _mphys_add_post_coupling_subsystems(self):
        self._mphys_check_coupling_order_inputs(self.options["post_coupling_order"])
        for discipline in self.options["post_coupling_order"]:
            self._mphys_add_post_coupling_subsystem_from_builder(
                discipline, self.options[f"{discipline}_builder"], self.name
            )

    def _mphys_initialize_builders(self):
        self.options["aero_builder"].initialize(self.comm)
        self.options["struct_builder"].initialize(self.comm)
        self.options["ldxfer_builder"].initialize(self.comm)

        geometry_builder = self.options["geometry_builder"]
        if geometry_builder is not None:
            geometry_builder.initialize(self.comm)

    def _mphys_add_mesh_and_geometry_subsystems(self):
        aero_builder = self.options["aero_builder"]
        struct_builder = self.options["struct_builder"]
        geometry_builder = self.options["geometry_builder"]

        if geometry_builder is None:
            self.mphys_add_subsystem(
                "aero_mesh", aero_builder.get_mesh_coordinate_subsystem(self.name)
            )
            self.mphys_add_subsystem(
                "struct_mesh", struct_builder.get_mesh_coordinate_subsystem(self.name)
            )
        else:
            self.add_subsystem(
                "aero_mesh", aero_builder.get_mesh_coordinate_subsystem(self.name)
            )
            self.add_subsystem(
                "struct_mesh", struct_builder.get_mesh_coordinate_subsystem(self.name)
            )
            self.mphys_add_subsystem(
                "geometry", geometry_builder.get_mesh_coordinate_subsystem(self.name)
            )
            self.connect("aero_mesh.x_aero0", "geometry.x_aero_in")
            self.connect("struct_mesh.x_struct0", "geometry.x_struct_in")



from .scenario import Scenario

class ScenarioStructural(Scenario):
    def initialize(self):
        """
        A class to perform a single discipline structural case
        for structural solvers which can compute their own set of loads.
        The Scenario will add the structural builder's precoupling subsystem,
        the coupling subsystem, and the postcoupling subsystem.
        """
        super().initialize()
        self.options.declare('struct_builder', recordable=False,
                             desc='The MPhys builder for the structural solver')
        self.options.declare('in_MultipointParallel', default=False, types=bool,
                             desc='Set to `True` if adding this scenario inside a MultipointParallel Group.')
        self.options.declare('geometry_builder', default=None, recordable=False,
                             desc='The optional MPhys builder for the geometry')

    def _mphys_scenario_setup(self):
        struct_builder = self.options['struct_builder']
        geometry_builder = self.options['geometry_builder']

        if self.options['in_MultipointParallel']:
            struct_builder.initialize(self.comm)

            if geometry_builder is not None:
                geometry_builder.initialize(self.comm)
                self.add_subsystem('mesh',struct_builder.get_mesh_coordinate_subsystem(self.name))
                self.mphys_add_subsystem('geometry',geometry_builder.get_mesh_coordinate_subsystem(self.name))
                self.connect('mesh.x_struct0','geometry.x_struct_in')
            else:
                self.mphys_add_subsystem('mesh',struct_builder.get_mesh_coordinate_subsystem(self.name))

        self._mphys_add_pre_coupling_subsystem_from_builder('struct', struct_builder, self.name)
        self.mphys_add_subsystem('coupling',struct_builder.get_coupling_group_subsystem(self.name))
        self._mphys_add_post_coupling_subsystem_from_builder('struct', struct_builder, self.name)



import numpy as np

import openmdao.api as om
from openmdao.utils.mpi import MPI


class DistributedVariableDescription:
    """
    Attributes of a variable for conversion from serial to distributed
    or distributed to serial
    """

    def __init__(self, name: str, shape: tuple, tags=[]):
        self.name = name
        self.shape = shape
        self.tags = tags


class DistributedConverter(om.ExplicitComponent):
    """
    An ExplicitComponent to convert from distributed to serial and serial to distributed variables.
    MPhys requires the coupling inputs and outputs to be distributed variables, so this
    class is provided to help with those conversions.
    For each mphys variable, a {variable}_serial version is created for the nonparallel solver to connect to and the
    distributed version will have the full vector on the root processor and zero length on the other processors.
    Given a list of distributed inputs in the options, the component will add variables to the inputs as distributed and
    produce {variable}_serial as outputs.
    Given a list of distributed outputs in the options, the component will add variables to the outputs as distributed and
    add {variable}_serial as inputs.

    """

    def initialize(self):
        self.options.declare(
            'distributed_inputs', default=[],
            desc='List of DistributedVariableDescription objects that will be converted from distributed to serial')
        self.options.declare(
            'distributed_outputs', default=[],
            desc='List of DistributedVariableDescription objects that will be converted from serial to distributed')

    def setup(self):
        for input in self.options['distributed_inputs']:
            self.add_input(input.name, shape_by_conn=True, tags=input.tags, distributed=True)
            self.add_output(f'{input.name}_serial', shape=input.shape, distributed=False)

        for output in self.options['distributed_outputs']:
            shape = output.shape if self.comm.Get_rank() == 0 else 0
            self.add_input(f'{output.name}_serial', shape_by_conn=True, distributed=False)
            self.add_output(output.name, shape=shape, tags=output.tags, distributed=True)

    def compute(self, inputs, outputs):
        for input in self.options['distributed_inputs']:
            if self.comm.Get_rank() == 0:
                outputs[f'{input.name}_serial'] = inputs[input.name]
            outputs[f'{input.name}_serial'] = self.comm.bcast(outputs[f'{input.name}_serial'])

        for output in self.options['distributed_outputs']:
            if self.comm.Get_rank() == 0:
                outputs[output.name] = inputs[f'{output.name}_serial']

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        if mode == 'fwd':
            for input in self.options['distributed_inputs']:
                if input.name in d_inputs and f'{input.name}_serial' in d_outputs:
                    if self.comm.Get_rank() == 0:
                        d_outputs[f'{input.name}_serial'] += d_inputs[input.name]
                    d_outputs[f'{input.name}_serial'] = self.comm.bcast(
                        d_outputs[f'{input.name}_serial'])

            for output in self.options['distributed_outputs']:
                if output.name in d_outputs and f'{output.name}_serial' in d_inputs:
                    if self.comm.Get_rank() == 0:
                        d_outputs[output.name] += d_inputs[f'{output.name}_serial']

        if mode == 'rev':
            for input in self.options['distributed_inputs']:
                if input.name in d_inputs and f'{input.name}_serial' in d_outputs:
                    if self.comm.Get_rank() == 0:
                        d_inputs[input.name] += d_outputs[f'{input.name}_serial']

            for output in self.options['distributed_outputs']:
                if output.name in d_outputs and f'{output.name}_serial' in d_inputs:
                    if self.comm.Get_rank() == 0:
                        d_inputs[f'{output.name}_serial'] += d_outputs[output.name]
                    d_inputs[f'{output.name}_serial'] = self.comm.bcast(
                        d_inputs[f'{output.name}_serial'])



class Builder:
    """
    MPHYS builder base class. Template for developers to create their builders.
    """

    def __init__(self):
        """
        Because the MPI communicator that will be used inside the OpenMDAO
        problem is not known when the builder is instantiated. The actual
        solver, transfer scheme, etc. should not be instantiated in the
        constructor.
        """
        pass

    def initialize(self, comm):
        """
        Initialize the solver, transfer scheme, etc.
        This method will be called when the MPI comm is available

        Parameters
        ----------
        comm : :class:`~mpi4py.MPI.Comm`
            The communicator object created for this xfer object instance.

        """
        pass

    def get_mesh_coordinate_subsystem(self, scenario_name=None):
        """
        The subsystem that contains the subsystem that will return the mesh
        coordinates

        Parameters
        ----------
        scenario_name : str or None
            The name of the scenario calling the builder.

        Returns
        -------
        mesh : :class:`~openmdao.api.Component` or :class:`~openmdao.api.Group` or None
            The openmdao subsystem that has an output of coordinates.
        """
        return None

    def get_coupling_group_subsystem(self, scenario_name=None):
        """
        The subsystem that this builder will add to the CouplingGroup

        Parameters
        ----------
        scenario_name : str or None
            The name of the scenario calling the builder.

        Returns
        -------
        subsystem : :class:`~openmdao.api.Component` or :class:`~openmdao.api.Group` or None
            The openmdao subsystem that handles all the computations for
            this solver. Transfer schemes can return multiple subsystems
        """
        return None

    def get_pre_coupling_subsystem(self, scenario_name=None):
        """
        Method that returns the openmdao subsystem to be added to each scenario before the coupling group

        Parameters
        ----------
        scenario_name : str or None
            The name of the scenario calling the builder.

        Returns
        -------
        subsystem : :class:`~openmdao.api.Component` or :class:`~openmdao.api.Group` or None
        """
        return None

    def get_post_coupling_subsystem(self, scenario_name=None):
        """
        Method that returns the openmdao subsystem to be added to each scenario after the coupling group

        Parameters
        ----------
        scenario_name : str or None
            The name of the scenario calling the builder.

        Returns
        -------
        subsystem : :class:`~openmdao.api.Component` or :class:`~openmdao.api.Group` or None
        """
        return None

    def get_number_of_nodes(self):
        """
        Method that returns the number of nodes defining the interface
        (input) mesh


        Returns
        -------
        number_of_nodes : int
            number of nodes in the computational domain
        """
        raise NotImplementedError('Function get_number_of_nodes has not been implemented in the builder.')

    def get_ndof(self):
        """
        The number of degrees of freedom used at each output location.

        Returns
        -------
        ndof : int
            number of degrees of freedom of each node in the domain
        """
        raise NotImplementedError('Function get_ndof has not been implemented in the builder.')

    def get_tagged_indices(self, tags):
        """
        Method that returns grid IDs for a list of body/boundary tags.

        Parameters
        ----------
        tags : list of integers or strings

        Returns
        -------
        grid_ids : list
            list of grid IDs that correspond to given body/boundary tags
        """
        raise NotImplementedError('Function get_tagged_indices has not been implemented in the builder.')



import openmdao.api as om

class SolverGroup(om.Group):

    def initialize(self):

        # define the inputs we need
        self.options.declare('builders', allow_none=False, recordable=False)
        self.options.declare('aero_discipline', allow_none=False)
        self.options.declare('struct_discipline', allow_none=False)
        self.options.declare('prop_discipline', allow_none=False)
        self.options.declare('as_coupling', allow_none=False)

    def setup(self):

        # set flags
        self.aero_discipline = self.options['aero_discipline']
        self.struct_discipline = self.options['struct_discipline']
        self.prop_discipline = self.options['prop_discipline']
        self.as_coupling = self.options['as_coupling']

        # set the builders
        self.aero_builder = self.options['builders']['aero']
        self.struct_builder = self.options['builders']['struct']
        self.prop_builder = self.options['builders']['prop']
        self.xfer_builder = self.options['builders']['xfer']

        # get the elements from each builder
        if self.aero_discipline:
            aero = self.aero_builder.get_element(as_coupling=self.as_coupling)
        if self.struct_discipline:
            struct = self.struct_builder.get_element(as_coupling=self.as_coupling)
        if self.as_coupling:
            disp_xfer, load_xfer = self.xfer_builder.get_element()

        if self.prop_discipline:
            prop = self.prop_builder.get_element()

        # add the subgroups
        if self.as_coupling:
            self.add_subsystem('disp_xfer', disp_xfer)
        if self.aero_discipline:
            self.add_subsystem('aero', aero)
        if self.as_coupling:
            self.add_subsystem('load_xfer', load_xfer)
        if self.struct_discipline:
            self.add_subsystem('struct', struct)
        if self.prop_discipline:
            self.add_subsystem('prop', prop)

        # set solvers
        # TODO add a nonlinear solver when we have feedback coupling to prop
        if self.as_coupling:
            self.nonlinear_solver=om.NonlinearBlockGS(maxiter=50, iprint=2, atol=1e-8, rtol=1e-8, use_aitken=True)
            self.linear_solver = om.LinearBlockGS(maxiter=50, iprint =2, atol=1e-8,rtol=1e-8)

    def configure(self):

        # do the connections, this can be also done in setup
        if self.as_coupling:
            self.connect('disp_xfer.u_aero', 'aero.u_aero')
            self.connect('aero.f_aero', 'load_xfer.f_aero')
            self.connect('load_xfer.f_struct', 'struct.f_struct')
            self.connect('struct.u_struct', ['disp_xfer.u_struct', 'load_xfer.u_struct'])



from .scenario import Scenario

class ScenarioAerodynamic(Scenario):
    def initialize(self):
        """
        A class to perform a single discipline aerodynamic case.
        The Scenario will add the aerodynamic builder's precoupling subsystem,
        the coupling subsystem, and the postcoupling subsystem.
        """
        super().initialize()

        self.options.declare('aero_builder', recordable=False,
                             desc='The MPhys builder for the aerodynamic solver')
        self.options.declare('in_MultipointParallel', default=False, types=bool,
                             desc='Set to `True` if adding this scenario inside a MultipointParallel Group.')
        self.options.declare('geometry_builder', default=None, recordable=False,
                             desc='The optional MPhys builder for the geometry')

    def _mphys_scenario_setup(self):
        aero_builder = self.options['aero_builder']
        geometry_builder = self.options['geometry_builder']

        if self.options['in_MultipointParallel']:
            aero_builder.initialize(self.comm)

            if geometry_builder is not None:
                geometry_builder.initialize(self.comm)
                self.add_subsystem('mesh',aero_builder.get_mesh_coordinate_subsystem(self.name))
                self.mphys_add_subsystem('geometry',geometry_builder.get_mesh_coordinate_subsystem(self.name))
                self.connect('mesh.x_aero0','geometry.x_aero_in')
            else:
                self.mphys_add_subsystem('mesh',aero_builder.get_mesh_coordinate_subsystem(self.name))
            self.connect('x_aero0','x_aero')

        self._mphys_add_pre_coupling_subsystem_from_builder('aero', aero_builder, self.name)
        self.mphys_add_subsystem('coupling',aero_builder.get_coupling_group_subsystem(self.name))
        self._mphys_add_post_coupling_subsystem_from_builder('aero', aero_builder, self.name)



import openmdao.api as om
from .coupling_group import CouplingGroup
from .geo_disp import GeoDisp


class CouplingAeroStructural(CouplingGroup):
    """
    The standard aerostructural coupling problem.
    """

    def initialize(self):
        self.options.declare('aero_builder', recordable=False)
        self.options.declare('struct_builder', recordable=False)
        self.options.declare('ldxfer_builder', recordable=False)
        self.options.declare("scenario_name", recordable=True, default=None)

    def setup(self):
        aero_builder = self.options['aero_builder']
        struct_builder = self.options['struct_builder']
        ldxfer_builder = self.options['ldxfer_builder']
        scenario_name = self.options['scenario_name']

        disp_xfer, load_xfer = ldxfer_builder.get_coupling_group_subsystem(scenario_name)
        aero = aero_builder.get_coupling_group_subsystem(scenario_name)
        struct = struct_builder.get_coupling_group_subsystem(scenario_name)

        geo_disp = GeoDisp(number_of_nodes=aero_builder.get_number_of_nodes())

        self.mphys_add_subsystem('disp_xfer', disp_xfer)
        self.mphys_add_subsystem('geo_disp', geo_disp)
        self.mphys_add_subsystem('aero', aero)
        self.mphys_add_subsystem('load_xfer', load_xfer)
        self.mphys_add_subsystem('struct', struct)

        self.nonlinear_solver = om.NonlinearBlockGS(maxiter=25, iprint=2,
                                                    atol=1e-8, rtol=1e-8,
                                                    use_aitken=True)
        self.linear_solver = om.LinearBlockGS(maxiter=25, iprint=2,
                                              atol=1e-8, rtol=1e-8,
                                              use_aitken=True)



import openmdao.api as om
from .scenario import Scenario
from .coupling_aerothermal import CouplingAeroThermal


class ScenarioAeroThermal(Scenario):

    def initialize(self):
        """
        A class to perform a single discipline aerodynamic case.
        The Scenario will add the aerodynamic builder's precoupling subsystem,
        the coupling subsystem, and the postcoupling subsystem.
        """
        super().initialize()
        
        self.options.declare('aero_builder', recordable=False,
                              desc='The Mphys builder for the aerodynamic solver')
        self.options.declare('thermal_builder', recordable=False,
                              desc='The Mphys builder for the Thermal solver')
        self.options.declare('thermalxfer_builder', recordable=False,
                              desc='The Mphys builder for the heat flux and temperature transfer')
        self.options.declare('in_MultipointParallel', default=False,
                             desc='Set to `True` if adding this scenario inside a MultipointParallel Group.')
        self.options.declare('geometry_builder', default=None, recordable=False,
                             desc='The optional Mphys builder for the geometry')

    def _mphys_scenario_setup(self):
        aero_builder = self.options['aero_builder']
        thermal_builder = self.options['thermal_builder']
        thermalxfer_builder = self.options['thermalxfer_builder']
        geometry_builder = self.options['geometry_builder']

        if self.options['in_MultipointParallel']:
            self._mphys_initialize_builders(aero_builder, thermal_builder,
                                            thermalxfer_builder, geometry_builder)
            self._mphys_add_mesh_and_geometry_subsystems(aero_builder, thermal_builder,
                                                         geometry_builder)

        self._mphys_add_pre_coupling_subsystem_from_builder('aero', aero_builder, self.name)
        self._mphys_add_pre_coupling_subsystem_from_builder('thermal', thermal_builder, self.name)
        self._mphys_add_pre_coupling_subsystem_from_builder('thermalxfer', thermalxfer_builder, self.name)

        coupling_group = CouplingAeroThermal(aero_builder=aero_builder,
                                                thermal_builder=thermal_builder,
                                                thermalxfer_builder=thermalxfer_builder,
                                                scenario_name=self.name)
        self.mphys_add_subsystem('coupling', coupling_group)

        self._mphys_add_post_coupling_subsystem_from_builder('thermalxfer', thermalxfer_builder, self.name)
        self._mphys_add_post_coupling_subsystem_from_builder('aero', aero_builder, self.name)
        self._mphys_add_post_coupling_subsystem_from_builder('thermal', thermal_builder, self.name)

    def _mphys_initialize_builders(self, aero_builder, thermal_builder,
                                   thermalxfer_builder, geometry_builder):
        aero_builder.initialize(self.comm)
        thermal_builder.initialize(self.comm)
        thermalxfer_builder.initialize(self.comm)
        if geometry_builder is not None:
            geometry_builder.initialize(self.comm)

    def _mphys_add_mesh_and_geometry_subsystems(self, aero_builder, thermal_builder,
                                                geometry_builder):

        if geometry_builder is None:
            self.mphys_add_subsystem('aero_mesh', aero_builder.get_mesh_coordinate_subsystem(self.name))
            self.mphys_add_subsystem('thermal_mesh', thermal_builder.get_mesh_coordinate_subsystem(self.name))
        else:
            self.add_subsystem('aero_mesh', aero_builder.get_mesh_coordinate_subsystem(self.name))
            self.add_subsystem('thermal_mesh', thermal_builder.get_mesh_coordinate_subsystem(self.name))
            self.mphys_add_subsystem('geometry', geometry_builder.get_mesh_coordinate_subsystem(self.name))
            self.connect('aero_mesh.x_aero0', 'geometry.x_aero_in')
            self.connect('thermal_mesh.x_thermal0', 'geometry.x_thermal_in')



import openmdao.api as om


def set_coupling_algorithms_in_scenarios(multipoint_group):
    """
    Set the stored linear and nonlinear solver into the coupling group if the
    scenario exists on this proc.
    Shared method between multipoint and multipoint parallel groups.
    Called during configure().
    """
    for scenario, solvers in multipoint_group.mphys_coupling_solvers:
        if solvers[0] is not None and scenario.comm:
            scenario.coupling.nonlinear_solver = solvers[0]

        if solvers[1] is not None and scenario.comm:
            scenario.coupling.linear_solver = solvers[1]


class Multipoint(om.Group):
    """
    An extension of the standard OpenMDAO group that adds the :func:`~mphys_add_scenario` method.
    For sequential evaluations of the MPhys scenarios.
    """

    def __init__(self, **kwargs):
        self.mphys_coupling_solvers = []
        super().__init__(**kwargs)

    def mphys_add_scenario(self, name, scenario, coupling_nonlinear_solver=None,
                           coupling_linear_solver=None):
        """
        Add an MPhys scenario

        Parameters
        ----------
        name : str
            The name of the scenario
        Scenario: :class:`~mphys.scenario.Scenario`
            The scenario object
        coupling_nonlinear_solver: openmdao.solvers.solver.NonlinearSolver
            The nonlinear solver to assign to the coupling group primal problem
        coupling_linear_solver: openmdao.solvers.solver.LinearSolver
            The linear solver to to assign to the coupling group sensitivity problem
        """
        solver_tuple = (coupling_nonlinear_solver, coupling_linear_solver)
        self.mphys_coupling_solvers.append((scenario, solver_tuple))
        return self.add_subsystem(name, scenario)

    def mphys_connect_scenario_coordinate_source(self, source, scenarios, disciplines):
        """
        A helper method to aid in connecting mesh coordinate sources to the scenarios
        in this multipoint group.
        The source and target variable names are assumed to be `'x_{discipline}0'`

        Parameters
        ----------
        source: openmdao.api.Group or openmdao.api.Component
            The source subsystem for the mesh coordinate variables
        scenarios : str or list[str]
            The names of the scenarios to be the target of the connections
        disciplines : str or list[str]
            The names of the disciplines for the mesh coordinates.

        """
        scenarios_list = scenarios if type(scenarios) == list else [scenarios]
        disciplines_list = disciplines if type(disciplines) == list else [disciplines]

        for scenario in scenarios_list:
            for discipline in disciplines_list:
                src = f'{source}.x_{discipline}0'
                target = f'{scenario}.x_{discipline}0'
                self.connect(src, target)
       

    def configure(self):
        return set_coupling_algorithms_in_scenarios(self)


class MultipointParallel(om.ParallelGroup):
    """
    An OpenMDAO parallel group that adds the :func:`~mphys_add_scenario` method.
    For simultaneous evaluations of the MPhys scenarios.
    """

    def __init__(self, **kwargs):
        self.mphys_coupling_solvers = []
        super().__init__(**kwargs)

    def mphys_add_scenario(self, name, scenario, coupling_nonlinear_solver=None,
                           coupling_linear_solver=None):
        """
        Add an MPhys scenario

        Parameters
        ----------
        name : str
            The name of the scenario
        Scenario: :class:`~mphys.scenario.Scenario`
            The scenario object
        coupling_nonlinear_solver: openmdao.solvers.solver.NonlinearSolver
            The nonlinear solver to assign to the coupling group primal problem
        coupling_linear_solver: openmdao.solvers.solver.LinearSolver
            The linear solver to to assign to the coupling group sensitivity problem
        """
        solver_tuple = (coupling_nonlinear_solver, coupling_linear_solver)
        self.mphys_coupling_solvers.append((scenario, solver_tuple))
        return self.add_subsystem(name, scenario)

    def configure(self):
        return set_coupling_algorithms_in_scenarios(self)



import os
from setuptools import setup, find_packages

__package_name__ = "mphys"
__package_version__ = "1.2.0"

mphys_root = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(mphys_root, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=__package_name__,
    version=__package_version__,
    description=("Components and related code for multiphysics"
        " problems in OpenMDAO"),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="",
    author_email="",
    zip_safe=False,
    packages = find_packages(),
    install_requires=[
          'numpy',
          'openmdao >= 3.25, != 3.27.0'
    ],
)



class MPHYS_Error(Exception):
    """
    Format the error message in a box to make it clear this
    was a expliclty raised exception.
    """
    def __init__(self, message):
        msg = '\n+'+'-'*78+'+'+'\n' + '| MPHYS Error: '
        i = 13
        for word in message.split():
            if len(word) + i + 1 > 78: # Finish line and start new one
                msg += ' '*(78-i)+'|\n| ' + word + ' '
                i = 1 + len(word)+1
            else:
                msg += word + ' '
                i += len(word)+1
        msg += ' '*(78-i) + '|\n' + '+'+'-'*78+'+'+'\n'
        print(msg)
        Exception.__init__(self)
